OrbitalDynamics

Struct OrbitalDynamics 

pub struct OrbitalDynamics {
    pub accel_models: Vec<Arc<dyn AccelModel + Sync>>,
}
OrbitalDynamics provides the equations of motion for any celestial dynamic, without state transition matrix computation.

Fields

accel_models: Vec<Arc<dyn AccelModel + Sync>>

Implementations

Source

impl OrbitalDynamics

pub fn point_masses(celestial_objects: Vec<i32>) -> Self

Initializes the point masses gravities with the provided list of bodies

Examples found in repository?
nyx-core/examples/03_geo_analysis/stationkeeping.rs (line 77)
28fn main() -> Result<(), Box<dyn Error>> {
29    pel::init();
30    // Set up the dynamics like in the orbit raise.
31    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
32    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
33
34    // Define the GEO orbit, and we're just going to maintain it very tightly.
35    let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
36    let orbit = Orbit::try_keplerian(42164.0, 1e-5, 0., 163.0, 75.0, 0.0, epoch, earth_j2000)?;
37    println!("{orbit:x}");
38
39    let sc = Spacecraft::builder()
40        .orbit(orbit)
41        .mass(Mass::from_dry_and_prop_masses(1000.0, 1000.0)) // 1000 kg of dry mass and prop, totalling 2.0 tons
42        .srp(SRPData::from_area(3.0 * 6.0)) // Assuming 1 kW/m^2 or 18 kW, giving a margin of 4.35 kW for on-propulsion consumption
43        .thruster(Thruster {
44            // "NEXT-STEP" row in Table 2
45            isp_s: 4435.0,
46            thrust_N: 0.472,
47        })
48        .mode(GuidanceMode::Thrust) // Start thrusting immediately.
49        .build();
50
51    // Set up the spacecraft dynamics like in the orbit raise example.
52
53    let prop_time = 30.0 * Unit::Day;
54
55    // Define the guidance law -- we're just using a Ruggiero controller as demonstrated in AAS-2004-5089.
56    let objectives = &[
57        Objective::within_tolerance(
58            StateParameter::Element(OrbitalElement::SemiMajorAxis),
59            42_165.0,
60            20.0,
61        ),
62        Objective::within_tolerance(
63            StateParameter::Element(OrbitalElement::Eccentricity),
64            0.001,
65            5e-5,
66        ),
67        Objective::within_tolerance(
68            StateParameter::Element(OrbitalElement::Inclination),
69            0.05,
70            1e-2,
71        ),
72    ];
73
74    let ruggiero_ctrl = Ruggiero::from_max_eclipse(objectives, sc, 0.2)?;
75    println!("{ruggiero_ctrl}");
76
77    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
78
79    let mut jgm3_meta = MetaFile {
80        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
81        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
82    };
83    jgm3_meta.process(true)?;
84
85    let harmonics = GravityField::new(GravityFieldData::from_cof(
86        &jgm3_meta.uri,
87        8,
88        8,
89        true,
90        almanac.frame_info(IAU_EARTH_FRAME)?,
91    )?);
92    orbital_dyn.accel_models.push(harmonics);
93
94    let srp_dyn = SolarPressure::default_flux(EARTH_J2000, &almanac)?;
95    let sc_dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn)
96        .with_guidance_law(ruggiero_ctrl.clone());
97
98    println!("{sc_dynamics}");
99
100    // Finally, let's use the Monte Carlo framework built into Nyx to propagate spacecraft.
101
102    // Let's start by defining the dispersion.
103    // The MultivariateNormal structure allows us to define the dispersions in any of the orbital parameters, but these are applied directly in the Cartesian state space.
104    // Note that additional validation on the MVN is in progress -- https://github.com/nyx-space/nyx/issues/339.
105    let mc_rv = MvnSpacecraft::new(
106        sc,
107        vec![StateDispersion::zero_mean(
108            StateParameter::Element(OrbitalElement::SemiMajorAxis),
109            3.0,
110        )],
111    )?;
112
113    let my_mc = MonteCarlo::new(
114        sc, // Nominal state
115        mc_rv,
116        "03_geo_sk".to_string(), // Scenario name
117        None, // No specific seed specified, so one will be drawn from the computer's entropy.
118    );
119
120    // Build the propagator setup.
121    let setup = Propagator::rk89(
122        sc_dynamics.clone(),
123        IntegratorOptions::builder()
124            .min_step(10.0_f64.seconds())
125            .error_ctrl(ErrorControl::RSSCartesianStep)
126            .build(),
127    );
128
129    let num_runs = 25;
130    let rslts = my_mc.run_until_epoch(setup, almanac.clone(), sc.epoch() + prop_time, num_runs);
131
132    assert_eq!(rslts.runs.len(), num_runs);
133
134    rslts.to_parquet("03_geo_sk.parquet", ExportCfg::default())?;
135
136    Ok(())
137}
More examples
Hide additional examples
nyx-core/examples/03_geo_analysis/raise.rs (line 95)
27fn main() -> Result<(), Box<dyn Error>> {
28    pel::init();
29
30    // Dynamics models require planetary constants and ephemerides to be defined.
31    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
32    // This will automatically download the DE440s planetary ephemeris,
33    // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
34    // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
35    // planetary constants kernels.
36    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
37    // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
38    // references to many functions.
39    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
40    // Fetch the EME2000 frame from the Almabac
41    let eme2k = almanac.frame_info(EARTH_J2000).unwrap();
42    // Define the orbit epoch
43    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
44
45    // Build the spacecraft itself.
46    // Using slide 6 of https://aerospace.org/sites/default/files/2018-11/Davis-Mayberry_HPSEP_11212018.pdf
47    // for the "next gen" SEP characteristics.
48
49    // GTO start
50    let orbit = Orbit::keplerian(24505.9, 0.725, 7.05, 0.0, 0.0, 0.0, epoch, eme2k);
51
52    let sc = Spacecraft::builder()
53        .orbit(orbit)
54        .mass(Mass::from_dry_and_prop_masses(1000.0, 1000.0)) // 1000 kg of dry mass and prop, totalling 2.0 tons
55        .srp(SRPData::from_area(3.0 * 6.0)) // Assuming 1 kW/m^2 or 18 kW, giving a margin of 4.35 kW for on-propulsion consumption
56        .thruster(Thruster {
57            // "NEXT-STEP" row in Table 2
58            isp_s: 4435.0,
59            thrust_N: 0.472,
60        })
61        .mode(GuidanceMode::Thrust) // Start thrusting immediately.
62        .build();
63
64    let prop_time = 180.0 * Unit::Day;
65
66    // Define the guidance law -- we're just using a Ruggiero controller as demonstrated in AAS-2004-5089.
67    let objectives = &[
68        Objective::within_tolerance(
69            StateParameter::Element(OrbitalElement::SemiMajorAxis),
70            42_165.0,
71            20.0,
72        ),
73        Objective::within_tolerance(
74            StateParameter::Element(OrbitalElement::Eccentricity),
75            0.001,
76            5e-5,
77        ),
78        Objective::within_tolerance(
79            StateParameter::Element(OrbitalElement::Inclination),
80            0.05,
81            1e-2,
82        ),
83    ];
84
85    // Ensure that we only thrust if we have more than 20% illumination.
86    let ruggiero_ctrl = Ruggiero::from_max_eclipse(objectives, sc, 0.2).unwrap();
87    println!("{ruggiero_ctrl}");
88
89    // Define the high fidelity dynamics
90
91    // Set up the spacecraft dynamics.
92
93    // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
94    // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
95    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
96
97    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
98    // We're using the JGM3 model here, which is the default in GMAT.
99    let mut jgm3_meta = MetaFile {
100        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
101        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
102    };
103    // And let's download it if we don't have it yet.
104    jgm3_meta.process(true)?;
105
106    // Build the spherical harmonics.
107    // The harmonics must be computed in the body fixed frame.
108    // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
109    let harmonics = GravityField::new(
110        GravityFieldData::from_cof(
111            &jgm3_meta.uri,
112            8,
113            8,
114            true,
115            almanac.frame_info(IAU_EARTH_FRAME)?,
116        )
117        .unwrap(),
118    );
119
120    // Include the spherical harmonics into the orbital dynamics.
121    orbital_dyn.accel_models.push(harmonics);
122
123    // We define the solar radiation pressure, using the default solar flux and accounting only
124    // for the eclipsing caused by the Earth.
125    let srp_dyn = SolarPressure::default_flux(EARTH_J2000, &almanac)?;
126
127    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
128    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
129    let sc_dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn)
130        .with_guidance_law(ruggiero_ctrl.clone());
131
132    println!("{orbit:x}");
133
134    // We specify a minimum step in the propagator because the Ruggiero control would otherwise drive this step very low.
135    let (final_state, traj) = Propagator::rk89(
136        sc_dynamics.clone(),
137        IntegratorOptions::builder()
138            .min_step(10.0_f64.seconds())
139            .error_ctrl(ErrorControl::RSSCartesianStep)
140            .build(),
141    )
142    .with(sc, almanac.clone())
143    .for_duration_with_traj(prop_time)?;
144
145    let prop_usage = sc.mass.prop_mass_kg - final_state.mass.prop_mass_kg;
146    println!("{:x}", final_state.orbit);
147    println!("prop usage: {prop_usage:.3} kg");
148
149    // Finally, export the results for analysis, including the penumbra percentage throughout the orbit raise.
150    traj.to_parquet("./03_geo_raise.parquet", ExportCfg::default())?;
151
152    for status_line in ruggiero_ctrl.status(&final_state) {
153        println!("{status_line}");
154    }
155
156    ruggiero_ctrl
157        .achieved(&final_state)
158        .expect("objective not achieved");
159
160    Ok(())
161}
nyx-core/examples/02_jwst_covar_monte_carlo/main.rs (line 99)
26fn main() -> Result<(), Box<dyn Error>> {
27    pel::init();
28    // Dynamics models require planetary constants and ephemerides to be defined.
29    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
30    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
31
32    // Download the regularly update of the James Webb Space Telescope reconstucted (or definitive) ephemeris.
33    // Refer to https://naif.jpl.nasa.gov/pub/naif/JWST/kernels/spk/aareadme.txt for details.
34    let mut latest_jwst_ephem = MetaFile {
35        uri: "https://naif.jpl.nasa.gov/pub/naif/JWST/kernels/spk/jwst_rec.bsp".to_string(),
36        crc32: None,
37    };
38    latest_jwst_ephem.process(true)?;
39
40    // Load this ephem in the general Almanac we're using for this analysis.
41    let almanac = Arc::new(
42        MetaAlmanac::latest()
43            .map_err(Box::new)?
44            .load_from_metafile(latest_jwst_ephem, true)?,
45    );
46
47    // By loading this ephemeris file in the ANISE GUI or ANISE CLI, we can find the NAIF ID of the JWST
48    // in the BSP. We need this ID in order to query the ephemeris.
49    const JWST_NAIF_ID: i32 = -170;
50    // Let's build a frame in the J2000 orientation centered on the JWST.
51    const JWST_J2000: Frame = Frame::from_ephem_j2000(JWST_NAIF_ID);
52
53    // Since the ephemeris file is updated regularly, we'll just grab the latest state in the ephem.
54    let (earliest_epoch, latest_epoch) = almanac.spk_domain(JWST_NAIF_ID)?;
55    println!("JWST defined from {earliest_epoch} to {latest_epoch}");
56    // Fetch the state, printing it in the Earth J2000 frame.
57    let jwst_orbit = almanac.transform(JWST_J2000, EARTH_J2000, latest_epoch, None)?;
58    println!("{jwst_orbit:x}");
59
60    // Build the spacecraft
61    // SRP area assumed to be the full sunshield and mass if 6200.0 kg, c.f. https://webb.nasa.gov/content/about/faqs/facts.html
62    // SRP Coefficient of reflectivity assumed to be that of Kapton, i.e. 2 - 0.44 = 1.56, table 1 from https://amostech.com/TechnicalPapers/2018/Poster/Bengtson.pdf
63    let jwst = Spacecraft::builder()
64        .orbit(jwst_orbit)
65        .srp(SRPData {
66            area_m2: 21.197 * 14.162,
67            coeff_reflectivity: 1.56,
68        })
69        .mass(Mass::from_dry_mass(6200.0))
70        .build();
71
72    // Build up the spacecraft uncertainty builder.
73    // We can use the spacecraft uncertainty structure to build this up.
74    // We start by specifying the nominal state (as defined above), then the uncertainty in position and velocity
75    // in the RIC frame. We could also specify the Cr, Cd, and mass uncertainties, but these aren't accounted for until
76    // Nyx can also estimate the deviation of the spacecraft parameters.
77    let jwst_uncertainty = SpacecraftUncertainty::builder()
78        .nominal(jwst)
79        .frame(LocalFrame::RIC)
80        .x_km(0.5)
81        .y_km(0.3)
82        .z_km(1.5)
83        .vx_km_s(1e-4)
84        .vy_km_s(0.6e-3)
85        .vz_km_s(3e-3)
86        .build();
87
88    println!("{jwst_uncertainty}");
89
90    // Build the Kalman filter estimate.
91    // Note that we could have used the KfEstimate structure directly (as seen throughout the OD integration tests)
92    // but this approach requires quite a bit more boilerplate code.
93    let jwst_estimate = jwst_uncertainty.to_estimate()?;
94
95    // Set up the spacecraft dynamics.
96    // We'll use the point masses of the Earth, Sun, Jupiter (barycenter, because it's in the DE440), and the Moon.
97    // We'll also enable solar radiation pressure since the James Webb has a huge and highly reflective sun shield.
98
99    let orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN, JUPITER_BARYCENTER]);
100    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
101
102    // Finalize setting up the dynamics.
103    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
104
105    // Build the propagator set up to use for the whole analysis.
106    let setup = Propagator::default(dynamics);
107
108    // All of the analysis will use this duration.
109    let prediction_duration = 6.5 * Unit::Day;
110
111    // === Covariance mapping ===
112    // For the covariance mapping / prediction, we'll use the common orbit determination approach.
113    // This is done by setting up a spacecraft Kalman filter OD process, and predicting for the analysis duration.
114
115    // Build the propagation instance for the OD process.
116    let odp = SpacecraftKalmanOD::new(
117        setup.clone(),
118        KalmanVariant::DeviationTracking,
119        None,
120        BTreeMap::new(),
121        almanac.clone(),
122    );
123
124    // The prediction step is 1 minute by default, configured in the OD process, i.e. how often we want to know the covariance.
125    assert_eq!(odp.max_step, 1_i64.minutes());
126    // Finally, predict, and export the trajectory with covariance to a parquet file.
127    let od_sol = odp.predict_for(jwst_estimate, prediction_duration)?;
128    od_sol.to_parquet("./02_jwst_covar_map.parquet", ExportCfg::default())?;
129
130    // === Monte Carlo framework ===
131    // Nyx comes with a complete multi-threaded Monte Carlo frame. It's blazing fast.
132
133    let my_mc = MonteCarlo::new(
134        jwst, // Nominal state
135        jwst_estimate.to_random_variable()?,
136        "02_jwst".to_string(), // Scenario name
137        None, // No specific seed specified, so one will be drawn from the computer's entropy.
138    );
139
140    let num_runs = 5_000;
141    let rslts = my_mc.run_until_epoch(
142        setup,
143        almanac.clone(),
144        jwst.epoch() + prediction_duration,
145        num_runs,
146    );
147
148    assert_eq!(rslts.runs.len(), num_runs);
149    // Finally, export these results, computing the eclipse percentage for all of these results.
150
151    rslts.to_parquet("02_jwst_monte_carlo.parquet", ExportCfg::default())?;
152
153    Ok(())
154}
nyx-core/examples/05_cislunar_spacecraft_link_od/main.rs (line 81)
34fn main() -> Result<(), Box<dyn Error>> {
35    pel::init();
36
37    // ====================== //
38    // === ALMANAC SET UP === //
39    // ====================== //
40
41    let manifest_dir = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
42
43    let out = manifest_dir.join("data/04_output/");
44
45    let almanac = Arc::new(
46        Almanac::new(
47            &manifest_dir
48                .join("data/01_planetary/pck08.pca")
49                .to_string_lossy(),
50        )
51        .unwrap()
52        .load(
53            &manifest_dir
54                .join("data/01_planetary/de440s.bsp")
55                .to_string_lossy(),
56        )
57        .unwrap(),
58    );
59
60    let eme2k = almanac.frame_info(EARTH_J2000).unwrap();
61    let moon_iau = almanac.frame_info(IAU_MOON_FRAME).unwrap();
62
63    let epoch = Epoch::from_gregorian_tai(2021, 5, 29, 19, 51, 16, 852_000);
64    let nrho = Orbit::cartesian(
65        166_473.631_302_239_7,
66        -274_715.487_253_382_7,
67        -211_233.210_176_686_7,
68        0.933_451_604_520_018_4,
69        0.436_775_046_841_900_9,
70        -0.082_211_021_250_348_95,
71        epoch,
72        eme2k,
73    );
74
75    let tx_nrho_sc = Spacecraft::from(nrho);
76
77    let state_luna = almanac.transform_to(nrho, MOON_J2000, None).unwrap();
78    println!("Start state (dynamics: Earth, Moon, Sun gravity):\n{state_luna}");
79
80    let bodies = vec![EARTH, SUN];
81    let dynamics = SpacecraftDynamics::new(OrbitalDynamics::point_masses(bodies));
82
83    let setup = Propagator::rk89(
84        dynamics,
85        IntegratorOptions::builder().max_step(0.5.minutes()).build(),
86    );
87
88    /* == Propagate the NRHO vehicle == */
89    let prop_time = 1.1 * state_luna.period().unwrap();
90
91    let (nrho_final, mut tx_traj) = setup
92        .with(tx_nrho_sc, almanac.clone())
93        .for_duration_with_traj(prop_time)
94        .unwrap();
95
96    tx_traj.name = Some("NRHO Tx SC".to_string());
97
98    println!("{tx_traj}");
99
100    /* == Propagate an LLO vehicle == */
101    let llo_orbit =
102        Orbit::try_keplerian_altitude(110.0, 1e-4, 90.0, 0.0, 0.0, 0.0, epoch, moon_iau).unwrap();
103
104    let llo_sc = Spacecraft::builder().orbit(llo_orbit).build();
105
106    let (_, llo_traj) = setup
107        .with(llo_sc, almanac.clone())
108        .until_epoch_with_traj(nrho_final.epoch())
109        .unwrap();
110
111    // Export the subset of the first two hours.
112    llo_traj
113        .clone()
114        .filter_by_offset(..2.hours())
115        .to_parquet_simple(out.join("05_caps_llo_truth.pq"))?;
116
117    /* == Setup the interlink == */
118
119    let mut measurement_types = IndexSet::new();
120    measurement_types.insert(MeasurementType::Range);
121    measurement_types.insert(MeasurementType::Doppler);
122
123    let mut stochastics = IndexMap::new();
124
125    let sa45_csac_allan_dev = 1e-11;
126
127    stochastics.insert(
128        MeasurementType::Range,
129        StochasticNoise::from_hardware_range_km(
130            sa45_csac_allan_dev,
131            10.0.seconds(),
132            link_specific::ChipRate::StandardT4B(),
133            link_specific::SN0::Average(),
134        ),
135    );
136
137    stochastics.insert(
138        MeasurementType::Doppler,
139        StochasticNoise::from_hardware_doppler_km_s(
140            sa45_csac_allan_dev,
141            10.0.seconds(),
142            link_specific::CarrierFreq::SBand(),
143            link_specific::CN0::Average(),
144        ),
145    );
146
147    let interlink = InterlinkTxSpacecraft {
148        traj: tx_traj,
149        measurement_types,
150        integration_time: None,
151        timestamp_noise_s: None,
152        ab_corr: Aberration::LT,
153        stochastic_noises: Some(stochastics),
154    };
155
156    // Devices are the transmitter, which is our NRHO vehicle.
157    let mut devices = BTreeMap::new();
158    devices.insert("NRHO Tx SC".to_string(), interlink);
159
160    let mut configs = BTreeMap::new();
161    configs.insert(
162        "NRHO Tx SC".to_string(),
163        TrkConfig::builder()
164            .strands(vec![Strand {
165                start: epoch,
166                end: nrho_final.epoch(),
167            }])
168            .build(),
169    );
170
171    let mut trk_sim =
172        TrackingArcSim::with_seed(devices.clone(), llo_traj.clone(), configs, 0).unwrap();
173    println!("{trk_sim}");
174
175    let trk_data = trk_sim.generate_measurements(&almanac).unwrap();
176    println!("{trk_data}");
177
178    trk_data
179        .to_parquet_simple(out.clone().join("nrho_interlink_msr.pq"))
180        .unwrap();
181
182    // Run a truth OD where we estimate the LLO position
183    let llo_uncertainty = SpacecraftUncertainty::builder()
184        .nominal(llo_sc)
185        .x_km(1.0)
186        .y_km(1.0)
187        .z_km(1.0)
188        .vx_km_s(1e-3)
189        .vy_km_s(1e-3)
190        .vz_km_s(1e-3)
191        .build();
192
193    let mut proc_devices = devices.clone();
194
195    // Define the initial estimate, randomized, seed for reproducibility
196    let mut initial_estimate = llo_uncertainty.to_estimate_randomized(Some(0)).unwrap();
197    // Inflate the covariance -- https://github.com/nyx-space/nyx/issues/339
198    initial_estimate.covar *= 2.5;
199
200    // Increase the noise in the devices to accept more measurements.
201
202    for link in proc_devices.values_mut() {
203        for noise in &mut link.stochastic_noises.as_mut().unwrap().values_mut() {
204            *noise.white_noise.as_mut().unwrap() *= 3.0;
205        }
206    }
207
208    let init_err = initial_estimate
209        .orbital_state()
210        .ric_difference(&llo_orbit)
211        .unwrap();
212
213    println!("initial estimate:\n{initial_estimate}");
214    println!("RIC errors = {init_err}",);
215
216    let odp = InterlinkKalmanOD::new(
217        setup.clone(),
218        KalmanVariant::ReferenceUpdate,
219        Some(SigmaRejection::default()),
220        proc_devices,
221        almanac.clone(),
222    );
223
224    // Shrink the data to process.
225    let arc = trk_data.filter_by_offset(..2.hours());
226
227    let od_sol = odp.process_arc(initial_estimate, &arc).unwrap();
228
229    println!("{od_sol}");
230
231    od_sol
232        .to_parquet(
233            out.join("05_caps_interlink_od_sol.pq"),
234            ExportCfg::default(),
235        )
236        .unwrap();
237
238    let od_traj = od_sol.to_traj().unwrap();
239
240    od_traj
241        .ric_diff_to_parquet(
242            &llo_traj,
243            out.join("05_caps_interlink_llo_est_error.pq"),
244            ExportCfg::default(),
245        )
246        .unwrap();
247
248    let final_est = od_sol.estimates.last().unwrap();
249    assert!(final_est.within_3sigma(), "should be within 3 sigma");
250
251    println!("ESTIMATE\n{final_est:x}\n");
252    let truth = llo_traj.at(final_est.epoch()).unwrap();
253    println!("TRUTH\n{truth:x}");
254
255    let final_err = truth
256        .orbit
257        .ric_difference(&final_est.orbital_state())
258        .unwrap();
259    println!("ERROR {final_err}");
260
261    // Build the residuals versus reference plot.
262    let rvr_sol = odp
263        .process_arc(initial_estimate, &arc.resid_vs_ref_check())
264        .unwrap();
265
266    rvr_sol
267        .to_parquet(
268            out.join("05_caps_interlink_resid_v_ref.pq"),
269            ExportCfg::default(),
270        )
271        .unwrap();
272
273    let final_rvr = rvr_sol.estimates.last().unwrap();
274
275    println!("RMAG error {:.3} m", final_err.rmag_km() * 1e3);
276    println!(
277        "Pure prop error {:.3} m",
278        final_rvr
279            .orbital_state()
280            .ric_difference(&final_est.orbital_state())
281            .unwrap()
282            .rmag_km()
283            * 1e3
284    );
285
286    Ok(())
287}
nyx-core/examples/06_lunar_orbit_determination/main.rs (line 90)
35fn main() -> Result<(), Box<dyn Error>> {
36    pel::init();
37
38    // ====================== //
39    // === ALMANAC SET UP === //
40    // ====================== //
41
42    // Dynamics models require planetary constants and ephemerides to be defined.
43    // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45    let data_folder: PathBuf = [
46        env!("CARGO_MANIFEST_DIR"),
47        "examples",
48        "06_lunar_orbit_determination",
49    ]
50    .iter()
51    .collect();
52
53    let meta = data_folder.join("metaalmanac.dhall");
54
55    // Load this ephem in the general Almanac we're using for this analysis.
56    let almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
57        .map_err(Box::new)?
58        .process(true)
59        .map_err(Box::new)?;
60
61    // Lock the almanac (an Arc is a read only structure).
62    let almanac = Arc::new(almanac);
63
64    // Build a nominal trajectory
65    // TODO: Switch this to a sequence once the OD over a spacecraft sequence is implemented.
66
67    let epoch = Epoch::from_gregorian_utc_at_noon(2024, 2, 29);
68    let moon_j2000 = almanac.frame_info(MOON_J2000)?;
69
70    // To build the trajectory we need to provide a spacecraft template.
71    let orbiter = Spacecraft::builder()
72        .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0))
73        .srp(SRPData {
74            area_m2: 3.9 * 2.7,
75            coeff_reflectivity: 0.96,
76        })
77        .orbit(Orbit::try_keplerian_altitude(
78            150.0, 0.00212, 33.6, 45.0, 45.0, 0.0, epoch, moon_j2000,
79        )?) // Setting a zero orbit here because it's just a template
80        .build();
81
82    // ========================== //
83    // === BUILD NOMINAL TRAJ === //
84    // ========================== //
85
86    // Set up the spacecraft dynamics.
87
88    // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
89    // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
90    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
91
92    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
93    // We're using the GRAIL JGGRX model.
94    let mut jggrx_meta = MetaFile {
95        uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
96        crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
97    };
98    // And let's download it if we don't have it yet.
99    jggrx_meta.process(true)?;
100
101    // Build the spherical harmonics.
102    // The harmonics must be computed in the body fixed frame.
103    // We're using the long term prediction of the Moon principal axes frame.
104    let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
105    let sph_harmonics = GravityField::new(GravityFieldData::from_shadr(
106        &jggrx_meta.uri,
107        80,
108        80,
109        true,
110        almanac.frame_info(moon_pa_frame)?,
111    )?);
112
113    // Include the spherical harmonics into the orbital dynamics.
114    orbital_dyn.accel_models.push(sph_harmonics);
115
116    // We define the solar radiation pressure, using the default solar flux and accounting only
117    // for the eclipsing caused by the Earth and Moon.
118    // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
119    let srp_dyn = SolarPressure::new(vec![MOON_J2000], &almanac)?;
120
121    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
122    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
123    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
124
125    println!("{dynamics}");
126
127    let setup = Propagator::rk89(dynamics.clone(), IntegratorOptions::default());
128
129    let truth_traj = setup
130        .with(orbiter, almanac.clone())
131        .for_duration_with_traj(Unit::Day * 2)?
132        .1;
133
134    // ==================== //
135    // === OD SIMULATOR === //
136    // ==================== //
137
138    // Load the Deep Space Network ground stations.
139    // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
140    let ground_station_file = data_folder.join("dsn-network.yaml");
141    let devices = GroundStation::load_named(ground_station_file)?;
142
143    let proc_devices = devices.clone();
144
145    // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
146    // Nyx can build a tracking schedule for you based on the first station with access.
147    let configs: BTreeMap<String, TrkConfig> =
148        TrkConfig::load_named(data_folder.join("tracking-cfg.yaml"))?;
149
150    // Build the tracking arc simulation to generate a "standard measurement".
151    let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
152        devices.clone(),
153        truth_traj.clone(),
154        configs,
155        123, // Set a seed for reproducibility
156    )?;
157
158    trk.build_schedule(&almanac)?;
159    let arc = trk.generate_measurements(&almanac)?;
160    // Save the simulated tracking data
161    arc.to_parquet_simple("./data/04_output/06_lunar_simulated_tracking.parquet")?;
162
163    // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
164    println!("{arc}");
165
166    // Now that we have simulated measurements, we'll run the orbit determination.
167
168    // ===================== //
169    // === OD ESTIMATION === //
170    // ===================== //
171
172    let sc = SpacecraftUncertainty::builder()
173        .nominal(orbiter)
174        .frame(LocalFrame::RIC)
175        .x_km(0.5)
176        .y_km(0.5)
177        .z_km(0.5)
178        .vx_km_s(5e-3)
179        .vy_km_s(5e-3)
180        .vz_km_s(5e-3)
181        .build();
182
183    // Build the filter initial estimate, which we will reuse in the filter.
184    let initial_estimate = sc.to_estimate()?;
185
186    println!("== FILTER STATE ==\n{orbiter:x}\n{initial_estimate}");
187
188    // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
189    let process_noise = ProcessNoise3D::from_velocity_km_s(
190        &[1e-14, 1e-14, 1e-14],
191        1 * Unit::Hour,
192        10 * Unit::Minute,
193        None,
194    );
195
196    println!("{process_noise}");
197
198    // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
199    let odp = SpacecraftKalmanScalarOD::new(
200        setup,
201        KalmanVariant::ReferenceUpdate,
202        Some(SigmaRejection::default()),
203        proc_devices,
204        almanac.clone(),
205    )
206    .with_process_noise(process_noise);
207
208    let od_sol = odp.process_arc(initial_estimate, &arc)?;
209
210    let final_est = od_sol.estimates.last().unwrap();
211
212    println!("{final_est}");
213
214    let ric_err = truth_traj
215        .at(final_est.epoch())?
216        .orbit
217        .ric_difference(&final_est.orbital_state())?;
218    println!("== RIC at end ==");
219    println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
220    println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
221
222    println!(
223        "Num residuals rejected: #{}",
224        od_sol.rejected_residuals().len()
225    );
226    println!(
227        "Percentage within +/-3: {}",
228        od_sol.residual_ratio_within_threshold(3.0).unwrap()
229    );
230    println!("Whitened residuals normal? {}", od_sol.is_normal(None)?);
231    println!("NIS test success? {}", od_sol.is_nis_consistent(None)?);
232
233    od_sol.to_parquet(
234        "./data/04_output/06_lunar_od_results.parquet",
235        ExportCfg::default(),
236    )?;
237
238    let od_trajectory = od_sol.to_traj()?;
239    // Build the RIC difference.
240    od_trajectory.ric_diff_to_parquet(
241        &truth_traj,
242        "./data/04_output/06_lunar_od_truth_error.parquet",
243        ExportCfg::default(),
244    )?;
245
246    Ok(())
247}
nyx-core/examples/03_geo_analysis/drift.rs (line 76)
26fn main() -> Result<(), Box<dyn Error>> {
27    pel::init();
28    // Dynamics models require planetary constants and ephemerides to be defined.
29    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
30    // This will automatically download the DE440s planetary ephemeris,
31    // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
32    // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
33    // planetary constants kernels.
34    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
35    // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
36    // references to many functions.
37    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
38    // Define the orbit epoch
39    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
40
41    // Define the orbit.
42    // First we need to fetch the Earth J2000 from information from the Almanac.
43    // This allows the frame to include the gravitational parameters and the shape of the Earth,
44    // defined as a tri-axial ellipoid. Note that this shape can be changed manually or in the Almanac
45    // by loading a different set of planetary constants.
46    let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
47
48    // Placing this GEO bird just above Colorado.
49    // In theory, the eccentricity is zero, but in practice, it's about 1e-5 to 1e-6 at best.
50    let orbit = Orbit::try_keplerian(42164.0, 1e-5, 0., 163.0, 75.0, 0.0, epoch, earth_j2000)?;
51    // Print in in Keplerian form.
52    println!("{orbit:x}");
53
54    let state_bf = almanac.transform_to(orbit, IAU_EARTH_FRAME, None)?;
55    let (orig_lat_deg, orig_long_deg, orig_alt_km) = state_bf.latlongalt()?;
56
57    // Nyx is used for high fidelity propagation, not Keplerian propagation as above.
58    // Nyx only propagates Spacecraft at the moment, which allows it to account for acceleration
59    // models such as solar radiation pressure.
60
61    // Let's build a cubesat sized spacecraft, with an SRP area of 10 cm^2 and a mass of 9.6 kg.
62    let sc = Spacecraft::builder()
63        .orbit(orbit)
64        .mass(Mass::from_dry_mass(9.60))
65        .srp(SRPData {
66            area_m2: 10e-4,
67            coeff_reflectivity: 1.1,
68        })
69        .build();
70    println!("{sc:x}");
71
72    // Set up the spacecraft dynamics.
73
74    // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
75    // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
76    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
77
78    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
79    // We're using the JGM3 model here, which is the default in GMAT.
80    let mut jgm3_meta = MetaFile {
81        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
82        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
83    };
84    // And let's download it if we don't have it yet.
85    jgm3_meta.process(true)?;
86
87    // Build the spherical harmonics.
88    // The harmonics must be computed in the body fixed frame.
89    // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
90    let harmonics_21x21 = GravityField::new(
91        GravityFieldData::from_cof(
92            &jgm3_meta.uri,
93            21,
94            21,
95            true,
96            almanac.frame_info(IAU_EARTH_FRAME)?,
97        )
98        .unwrap(),
99    );
100
101    // Include the spherical harmonics into the orbital dynamics.
102    orbital_dyn.accel_models.push(harmonics_21x21);
103
104    // We define the solar radiation pressure, using the default solar flux and accounting only
105    // for the eclipsing caused by the Earth and Moon.
106    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
107
108    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
109    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
110    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
111
112    println!("{dynamics}");
113
114    // Finally, let's propagate this orbit to the same epoch as above.
115    // The first returned value is the spacecraft state at the final epoch.
116    // The second value is the full trajectory where the step size is variable step used by the propagator.
117    let (future_sc, trajectory) = Propagator::default(dynamics)
118        .with(sc, almanac.clone())
119        .until_epoch_with_traj(epoch + Unit::Century * 0.03)?;
120
121    println!("=== High fidelity propagation ===");
122    println!(
123        "SMA changed by {:.3} km",
124        orbit.sma_km()? - future_sc.orbit.sma_km()?
125    );
126    println!(
127        "ECC changed by {:.6}",
128        orbit.ecc()? - future_sc.orbit.ecc()?
129    );
130    println!(
131        "INC changed by {:.3e} deg",
132        orbit.inc_deg()? - future_sc.orbit.inc_deg()?
133    );
134    println!(
135        "RAAN changed by {:.3} deg",
136        orbit.raan_deg()? - future_sc.orbit.raan_deg()?
137    );
138    println!(
139        "AOP changed by {:.3} deg",
140        orbit.aop_deg()? - future_sc.orbit.aop_deg()?
141    );
142    println!(
143        "TA changed by {:.3} deg",
144        orbit.ta_deg()? - future_sc.orbit.ta_deg()?
145    );
146
147    // We also have access to the full trajectory throughout the propagation.
148    println!("{trajectory}");
149
150    println!("Spacecraft params after 3 years without active control:\n{future_sc:x}");
151
152    // With the trajectory, let's build a few data products.
153
154    // 1. Export the trajectory as a parquet file, which includes the Keplerian orbital elements.
155
156    let analysis_step = Unit::Minute * 5;
157
158    trajectory.to_parquet(
159        "./03_geo_hf_prop.parquet",
160        ExportCfg::builder().step(analysis_step).build(),
161    )?;
162
163    // 2. Compute the latitude, longitude, and altitude throughout the trajectory by rotating the spacecraft position into the Earth body fixed frame.
164
165    // We iterate over the trajectory, grabbing a state every two minutes.
166    let mut offset_s = vec![];
167    let mut epoch_str = vec![];
168    let mut longitude_deg = vec![];
169    let mut latitude_deg = vec![];
170    let mut altitude_km = vec![];
171
172    for state in trajectory.every(analysis_step) {
173        // Convert the GEO bird state into the body fixed frame, and keep track of its latitude, longitude, and altitude.
174        // These define the GEO stationkeeping box.
175
176        let this_epoch = state.epoch();
177
178        offset_s.push((this_epoch - orbit.epoch).to_seconds());
179        epoch_str.push(this_epoch.to_isoformat());
180
181        let state_bf = almanac.transform_to(state.orbit, IAU_EARTH_FRAME, None)?;
182        let (lat_deg, long_deg, alt_km) = state_bf.latlongalt()?;
183        longitude_deg.push(long_deg);
184        latitude_deg.push(lat_deg);
185        altitude_km.push(alt_km);
186    }
187
188    println!(
189        "Longitude changed by {:.3} deg -- Box is 0.1 deg E-W",
190        orig_long_deg - longitude_deg.last().unwrap()
191    );
192
193    println!(
194        "Latitude changed by {:.3} deg -- Box is 0.05 deg N-S",
195        orig_lat_deg - latitude_deg.last().unwrap()
196    );
197
198    println!(
199        "Altitude changed by {:.3} km -- Box is 30 km",
200        orig_alt_km - altitude_km.last().unwrap()
201    );
202
203    // Build the station keeping data frame.
204    let mut sk_df = df!(
205        "Offset (s)" => offset_s.clone(),
206        "Epoch (UTC)" => epoch_str.clone(),
207        "Longitude E-W (deg)" => longitude_deg,
208        "Latitude N-S (deg)" => latitude_deg,
209        "Altitude (km)" => altitude_km,
210
211    )?;
212
213    // Create a file to write the Parquet to
214    let file = File::create("./03_geo_lla.parquet").expect("Could not create file");
215
216    // Create a ParquetWriter and write the DataFrame to the file
217    ParquetWriter::new(file).finish(&mut sk_df)?;
218
219    Ok(())
220}
pub fn two_body() -> Self

Initializes a OrbitalDynamics which does not simulate the gravity pull of other celestial objects but the primary one.

pub fn new(accel_models: Vec<Arc<dyn AccelModel + Sync>>) -> Self

Initialize orbital dynamics with a list of acceleration models

pub fn from_model(accel_model: Arc<dyn AccelModel + Sync>) -> Self

Initialize new orbital mechanics with the provided model. Note: Orbital dynamics always include two body dynamics, these cannot be turned off.

Source

impl OrbitalDynamics

pub fn dual_eom( &self, _delta_t_s: f64, osc: &Orbit, almanac: &Almanac, ) -> Result<(Vector6<f64>, Matrix6<f64>), DynamicsError>