pub struct PointMasses {
pub celestial_objects: Vec<i32>,
pub correction: Option<Aberration>,
}Expand description
PointMasses model
Fields
celestial_objects: Vec<i32>correction: Option<Aberration>Light-time correction computation if extra point masses are needed
Implementations
Sourceimpl PointMasses
impl PointMasses
Sourcepub fn new(celestial_objects: Vec<i32>) -> Arc<Self>
pub fn new(celestial_objects: Vec<i32>) -> Arc<Self>
Initializes the point masses gravities with the provided list of bodies
Sourcepub fn with_correction(
celestial_objects: Vec<i32>,
correction: Option<Aberration>,
) -> Self
pub fn with_correction( celestial_objects: Vec<i32>, correction: Option<Aberration>, ) -> Self
Initializes the point masses gravities with the provided list of bodies, and accounting for some light time correction